#ifndef _DISTREE_VERTEX_H_
#define _DISTREE_VERTEX_H_
#pragma once
#include "vec3f.h"
class distree_vertex_t
{
public:
    int       key;
    union{
        struct{
            int node_index;//the index of the node that is incident to the vertex. Only used for removing duplicates
            int node_local_id;//[0:7] indicate which vertex of the node. Only used for removing duplicates
        };
        struct{
            float  eucl_dist; //minimal euclidean distance from vertex to the nearest feature.
            float  maxnorm_dist; //minimal maxnorm distance from vertex to the nearest feature.
        };
    };
    lily::vec3f    coords;
    int      min_dist_type;
    //-1: There is no distance computed for this vertex, possibly because there is no local features.
    //0: min distance is between the vertex and the features of its direct octant
    //1: min distance is between the vertex and the features of its parent octant
    //2: min distance is between the vertex and the features of its grandparent octant
    //-1: invalid value(default)
    //int       min_node_index;//the index of the node that shares this vertex when local min dist is computed.
    int      is_send;//label that denotes if this vertex has been sent to remote PE before.
public:
    distree_vertex_t(void):node_index(0),node_local_id(0),
        coords(0.0f),min_dist_type(-1),is_send(-1){};
    ~distree_vertex_t(void){};
    //functor used for comparison
    bool operator()(const distree_vertex_t& a,const distree_vertex_t& b)
    {
        return a.key < b.key;
    }
};
#endif

