#include <mpi.h>
#include "vec3i.h"
#include "DError.h"
#include "dvolume.h"
#include "dutility.h"
#include "constant.h"
#include "distree_t.h"
#include "utility.h"
using namespace lily;

volume_t::volume_t( int size_x, int size_y, int size_z, float isovalue )
    :d_size(size_x, size_y, size_z),m_isovalue(isovalue)
{
    int t_size = d_size[0] * d_size[1] * d_size[2];
    m_data.resize(t_size,0.0f);
}

volume_t::volume_t(const vec3i &size3d, float isovalue )
    :d_size(size3d),m_isovalue(isovalue)
{
    int t_size = d_size[0] * d_size[1] * d_size[2];
    m_data.resize(t_size,0.0f);
}
void volume_t::load_file(const char* file_path)
{
    FILE* fp = fopen(file_path, "r");
    if (!fp)
    {
        fprintf(stderr, "volume file open failed!\n");
        exit(1);
    }

    size_t count = d_size[0];
    count *= d_size[1];
    count *= d_size[2];
    fread(m_data.data_ptr(), sizeof(unsigned char), count, fp);

    fclose(fp);
}
volume_t::~volume_t()
{
}

void volume_t::set_volume_value( int x, int y, int z, float v )
{
    int id = 0;
        D_ASSERT(x >= 0 && x < d_size[0] &&
                 y >= 0 && y < d_size[1] &&
                 z >= 0 && z < d_size[2],
                 "set_volume_value(): Over flow",
                 DERROR_ERR_MEM);
    id = x + y * d_size[0] + z * d_size[0] * d_size[1];
    m_data[id] = v;
}
float volume_t::get_volume_value( int x, int y, int z )
{
    D_ASSERT(x >= 0 && x < d_size[0] &&
             y >= 0 && y < d_size[1] &&
             z >= 0 && z < d_size[2],
             "Over flow",
             DERROR_ERR_MEM);
    int id = x + y * d_size[0] + z * d_size[0] * d_size[1];

    return m_data[id];
}

void volume_t::initial_volume(
    char*   filename,  color_t &mycolor,vec3i &global_feature_size,
    int     groupsize, vec3i &np_xyz, int myid )
{
    if (filename){
        readVolumeFromFile(filename, global_feature_size, mycolor, groupsize, np_xyz, myid);
    }
    else{
        regularSample( mycolor, groupsize, myid, 4);
    }
}
void volume_t::save_volume( char *filename )
{
    char name[1024];
    FILE *fp = 0;

    sprintf(name, "%s.dat", filename);
    fp = fopen(name, "wb");
    D_ASSERT(fp, "Can not open file", DERROR_ERR_IO);
    fwrite(&(m_data[0]), sizeof(float), m_data.size(), fp);
    fclose(fp);

    sprintf(name, "%s.hdr", filename);
    fp = fopen(name, "w");
    D_ASSERT(fp, "Can not open file", DERROR_ERR_IO);
    fprintf(fp, "%d %d %d\nFLOAT\n",
            d_size[0],
            d_size[1],
            d_size[2]);
    fclose(fp);
}

void volume_t::randomSample(color_t &mycolor, int groupsize, int myid, int mod)
{
    int randnum;
    int test_count = 10, i = 0;
    vec3i local_size(d_size);

    srand(myid);
    randnum = rand();
    if (( randnum % groupsize) < mod) {
        fprintf(stderr, "%d initialize volume\n", myid);
        fflush(stderr);
        for (i = 0; i < test_count; i++) {
            set_volume_value(
                (int)(local_size.x() * ((float)(rand()) / RAND_MAX)),
                (int)(local_size.y() * ((float)(rand()) / RAND_MAX)),
                (int)(local_size.z() * ((float)(rand()) / RAND_MAX)),
                1.0f);
        }

        mycolor = BLACK;
    } else {
        mycolor = WHITE;
    }
}

void volume_t::fixedSample(color_t &mycolor, int groupsize, int myid)
{
    vec3i local_size(d_size);
    if (myid == 0/*myid*/ ){
        set_volume_value(
            (int)((local_size.x()/*-1*/) * 0.5f),
            (int)((local_size.y()/*-1*/) * 0.5f),
            (int)((local_size.z()/*-1*/) * 0.5f),
            m_isovalue);
        mycolor = BLACK;
    }
    else if(myid == 1){
        mycolor = WHITE;
        return;
        set_volume_value(
            (int)((local_size.x()-1) * 0.5f),
            (int)((local_size.y()-1) * 0.5f),
            (int)((local_size.z()-1) * 0.5f),
            m_isovalue);
        mycolor = BLACK;
    }
    else{
        mycolor = WHITE;
    }
}

void volume_t::regularSample(color_t &mycolor,int groupsize, int myid, int mod)
{
    int block_low_x,  block_low_y,  block_low_z;
    int x, y, z;

    block_low_x  = (int)m_bound.pMin.x();
    block_low_y  = (int)m_bound.pMin.y();
    block_low_z  = (int)m_bound.pMin.z();
    mycolor = WHITE;
    for (z = 0; z < d_size[2] ; z++){
        for (y = 0; y < d_size[1] ; y++){
            for (x = 0; x < d_size[0]; x++){
                if (  (x+block_low_x)%mod==0
                    &&(y+block_low_y)%mod==0
                    &&(z+block_low_z)%mod==0){
                    set_volume_value(x, y, z, m_isovalue);
                    mycolor = BLACK;
                }
            }
        }
    }
}

void volume_t::readVolumeFromFile(const char* file_name, vec3i& volumeDataSize,
    color_t &mycolor, int groupsize, vec3i &np_xyz, int myid )
{
    MPI_File fh;
    int gsizes[3];
    int psizes[3];
    int lsizes[3];
    int start_indices[3];
    int   local_array_size;
    char  err_msg[128];
    float *local_array = NULL;
    MPI_Datatype filetype;
    MPI_File_open(MPI_COMM_WORLD, const_cast<char*>(file_name), MPI_MODE_RDONLY
                  ,MPI_INFO_NULL, &fh);

    gsizes[0] = volumeDataSize.x();
    gsizes[1] = volumeDataSize.y();
    gsizes[2] = volumeDataSize.z();

    psizes[0] = np_xyz.x();
    psizes[1] = np_xyz.y();
    psizes[2] = np_xyz.z();

    lsizes[0] = d_size[0];
    lsizes[1] = d_size[1];
    lsizes[2] = d_size[2];

    start_indices[0] = (int)m_bound.pMin.x();
    start_indices[1] = (int)m_bound.pMin.y();
    start_indices[2] = (int)m_bound.pMin.z();
    D_ASSERT_P(myid, groupsize == psizes[0] * psizes[1] * psizes[2],
                "groupsize unmatched!",DERROR_ERR_MEM);
    sprintf(err_msg, "Collective IO Exception: gsizes[%d,%d,%d]= < lsizes[%d,%d,%d]=\n",
        gsizes[0],gsizes[1],gsizes[2], lsizes[0], lsizes[1], lsizes[2]);

    D_ASSERT_P(myid, gsizes[0]>=lsizes[0] && gsizes[1] >= lsizes[1]
                && gsizes[2]>=lsizes[2], err_msg,DERROR_ERR_IO);

    local_array = data_ptr(m_data);

    local_array_size = lsizes[0] * lsizes[1] * lsizes[2];

    MPI_Type_create_subarray(3, gsizes, lsizes, start_indices, MPI_ORDER_FORTRAN,
                             MPI_FLOAT, &filetype);

    MPI_Type_commit(&filetype);

    MPI_File_set_view(fh, 0, MPI_FLOAT, filetype, "native", MPI_INFO_NULL);
    MPI_File_read_all(fh, local_array, local_array_size, MPI_FLOAT, MPI_STATUS_IGNORE);

    MPI_File_close(&fh);

    mycolor = BLACK;

}

void volume_t::initial_distree_feature_array(distree_t &tree )
{
    int isocount = 0;
    int nx = getDimension()[0];
    int ny = getDimension()[1];
    int nz = getDimension()[2];
    int ix, iy, iz;
    float val;
    D_ASSERT_P(tree.get_processor_id(), m_data.size()>0, "initial_distree_feature_array(): local volume data size is 0", DERROR_ERR_MEM);

    vec3f block_offset = m_bound.pMin;
    for (iz = 0; iz < nz ; iz++){
        for (iy = 0; iy < ny ; iy++){
            for (ix = 0; ix < nx ; ix++){
                val = m_data[iz*ny*nx + iy*nx + ix];

                if (fabsf(val - m_isovalue) < FLOAT_EPSILON) {
                    isocount++;
                    m_isovoxels.push_back(voxel_t(
                                            vec3f((float)ix,(float)iy,(float)iz)+block_offset,
                                            m_isovalue
                                            ));
                }
            }
        }
    }
    m_isovoxels.shrink_to_fit();
}

size_t volume_t::size_feature() const
{
    return m_isovoxels.size();
}

bool volume_t::collect_feature( size_t child_node_index,
                                size_t parent_node_index,
                                distree_t& tree, bool is_profiling/*=false */ )
{
    distree_node_t& parent = tree.m_node_array[parent_node_index];
    const svector<int>& parent_feature_indices =
                        parent.triple_feature_indices;
    size_t parent_feature_count = parent_feature_indices.size();

    distree_node_t& cur_child = tree.m_node_array[child_node_index];
#ifdef USE_MAX_NORM
    geometry::BBox child_3x_bbox = cur_child.m_bound;
    child_3x_bbox.Expand(child_3x_bbox.MaximumExtentf());
#endif
    float child_radius = cur_child.m_bound.BoxRadius();
    int   child_id = tree.compute_local_child_key(cur_child);

    bool is_leaf = true;

    for (size_t i = 0; i <  parent_feature_count; i++) {
        D_ASSERT_P(tree.m_myid, parent_feature_indices[i] < size_feature(),
            "Memory overflow", DERROR_ERR_MEM);
        voxel_t& cur_voxel = m_isovoxels[parent_feature_indices[i]];
#ifdef USE_MAX_NORM
        if (child_3x_bbox.Inside(cur_voxel.coords))
#else
        if(POINT_INSIDE_SPHERE(cur_child.m_bound.Center(),
            child_radius * 3/*+0.1f*/, cur_voxel.coords))
#endif
        {
            cur_child.triple_feature_indices.push_back(parent_feature_indices[i]);
        }
        
    }
    is_leaf = cur_child.triple_feature_indices.empty();
    if (   cur_child.m_bound.MaximumExtentf() < tree.min_extend*2.0f
        || parent.level == tree.max_depth_id-1){
            is_leaf = DTRUE;
    }

    D_ASSERT_P(tree.get_processor_id(),
        cur_child.triple_feature_indices.size() <= parent_feature_indices.size(),
        "Memory overflow",
        DERROR_ERR_MEM);
    cur_child.triple_feature_indices.shrink_to_fit();
    return is_leaf;
}

float volume_t::compute_vertex_feature_distance( distree_vertex_t& cur_vertex,
                                                 int feature_index, distree_t& tree )
{
    D_ASSERT_P(tree.m_myid, feature_index < m_isovoxels.size(),
                "compute_vertex_feature_distance(): feature index out of bounds", DERROR_ERR_MEM);
    voxel_t &voxel = m_isovoxels[feature_index];
    return (cur_vertex.coords - voxel.coords).length();
}

float volume_t::compute_vertex_feature_distance( vec3f& coords,
                                                 int feature_index, distree_t& tree )
{
    D_ASSERT_P(tree.m_myid, feature_index < m_isovoxels.size(),
        "compute_vertex_feature_distance(): feature index out of bounds", DERROR_ERR_MEM);
    voxel_t &voxel = m_isovoxels[feature_index];
    return (coords - voxel.coords).length();
}
