#ifndef DVOLUME_H
#define DVOLUME_H

#include <stdint.h>
#include "vec3f.h"
#include "vec3i.h"
#include "BBox.h"
#include "dutility.h"
#include "IFeatures.h"
#include "bitmap_t.h"
#include "utility.h"

class voxel_t{
public:
    lily::vec3f   coords;
    float   val;
    int     key;
public:
    voxel_t(const lily::vec3f &pos=lily::vec3f(0.0f), float v=0.0f, int k=0)
        :coords(pos),val(v),key(k){};
};

class volume_t :public IFeatures{
public:
    lily::BBox  m_bound;//the bbox of isoblock
    lily::vec3i d_size;//
    //int    t_size;
    svector<float>   m_data;
    float            m_isovalue;
    svector<voxel_t> m_isovoxels;
public:
    volume_t(int size_x, int size_y, int size_z, float isovalue=0.0f);
    volume_t(const lily::vec3i &size3d, float isovalue=0.0f);
    ~volume_t();

    void load_file(const char* file_path);
    virtual void initial_distree_feature_array(distree_t &tree );
    virtual size_t size_feature() const;
    virtual void*  data(){ return m_isovoxels.data_ptr();};
    virtual float compute_vertex_feature_distance( distree_vertex_t& cur_vertex,
                                                    int feature_index, distree_t& tree );
    virtual float compute_vertex_feature_distance( lily::vec3f& coords,
                                                    int feature_index, distree_t& tree );
    void  setIsovalue(float v){ m_isovalue = v;}
    float getIsovalue() const { return m_isovalue;}
    void  setBoundBox(const lily::BBox& box){ m_bound = box;};
    void  setBoundBox(const lily::vec3f& pmin, const lily::vec3f& pmax){
            m_bound.pMin=pmin; m_bound.pMax=pmax;
    }
    lily::vec3i getDimension() const{ return d_size;};
    //float *data_ptr(){ return m_data;}
    void  set_volume_value(int x, int y, int z, float v);

    float get_volume_value(int x, int y, int z);

    void  save_volume(char *filename);
    void  initial_volume( char* filename, color_t &mycolor,
                        lily::vec3i &global_feature_size, int groupsize,
                        lily::vec3i &np_xyz, int myid );

protected:
    void regularSample(color_t &mycolor,int groupsize, int myid, int mod);
    //file_name: directory of the original volume data file.
    //volumeDataSize: size of the original volume in the data file to be read.
    void readVolumeFromFile(const char* file_name, lily::vec3i& volumeDataSize, color_t &mycolor,
                                int groupsize, lily::vec3i &np_xyz, int myid );
    void randomSample(color_t &mycolor, int groupsize, int myid, int mod);
    void fixedSample(color_t &mycolor, int groupsize, int myid);

//     virtual void inherit_features(        distree_node_t& cur_child,
//                                           distree_node_t& parent,
//                                     const distree_t& tree );

    virtual bool collect_feature( size_t cur_node_index,
                                  size_t parent_node_index,
                                  distree_t& tree, bool is_profiling=false );
};

#endif

