#include "BBox.h"
#include "feature_geometry_t.h"
#include "constant.h"
#include "distree_t.h"
#include "obj_model_t.h"

using namespace lily;

feature_geometry_t::feature_geometry_t(void)
    :m_pobj_model(NULL)
{
}

feature_geometry_t::feature_geometry_t( obj_model_t* pobj_model )
    :m_pobj_model(pobj_model)
{
}


feature_geometry_t::~feature_geometry_t(void)
{
}

void feature_geometry_t::load_geometry_from_binfile( char* filename)
{
    D_ASSERT_P(0,m_pobj_model, "m_pobj_model is NULL, you probably forget to initialize it!\n",
                DERROR_ERR_MEM);
    m_pobj_model->obj_model_load_file(filename);
    //m_pobj_model->obj_model_remove_duplicated_vertex();
    //m_pobj_model->obj_model_remove_degenerated_triangle();
}
void feature_geometry_t::remove_degenerated_triangle()
{
    m_pobj_model->obj_model_remove_duplicated_vertex();
    m_pobj_model->obj_model_remove_degenerated_triangle();
}
void feature_geometry_t::initial_distree_feature_array( distree_t &tree )
{//Select the triangle geometry that falls into the assigned local region bbox.
    //int triangle_index_count = 0;
    int ntriangles = m_pobj_model->triangle_num;
    //svector<int> triangle_index_array;
    //triangle_index_array.reserve(ntriangles);
    //computing bbox of distree assigned local region
    //int local_leaf_num = tree.m_assigned_local_leaf_indices.size();
    vec3i pxyz = vec3i::index1D_to_index3D(tree.m_myid,tree.m_np_xyz);
    vec3f global_bbox(m_pobj_model->m_local_bbox.getDimension());//global_feature_size(global_domain_size);//
    vec3f local_block_min = vec3f(BLOCK_LOW_FLOAT3D(pxyz, tree.m_np_xyz, global_bbox))
                                    +m_pobj_model->m_local_bbox.pMin;
    vec3f local_block_max = vec3f(BLOCK_HIGH_FLOAT3D(pxyz, tree.m_np_xyz, global_bbox))
                                    +m_pobj_model->m_local_bbox.pMin;
    BBox local_region_bbox(local_block_min, local_block_max);

    //Before moving ahead, make sure the geometry coordinates is computed
    //relative to its obj_bbox.pMin, not global_domain_bbox.pMin. Because
    //we use obj_bbox.pMin as the origin in the entire process.
    triangle_t cur_triangle;
    m_features_triangle.reserve(ntriangles);

    for (int iTria = 0; iTria < ntriangles; iTria++)
    {
        bool isNormal = m_pobj_model->obj_model_get_triangle_at(iTria, cur_triangle);
        if (isNormal)
        {
            if(    local_region_bbox.Overlap(cur_triangle.v[0])
                || local_region_bbox.Overlap(cur_triangle.v[1])
                || local_region_bbox.Overlap(cur_triangle.v[2])
              )
            {
                m_features_triangle.push_back(cur_triangle);
            }
        }
    }

    m_features_triangle.shrink_to_fit();
    m_pobj_model->obj_free();
}

size_t feature_geometry_t::size_feature() const
{
    return m_features_triangle.size();
}

void* feature_geometry_t::data()
{
    return m_features_triangle.data_ptr();
}

bool feature_geometry_t::collect_feature( size_t child_node_index,
                                          size_t parent_node_index,
                                          distree_t& tree, bool is_profiling/*=false */ )
{
    distree_node_t& parent = tree.m_node_array[parent_node_index];
    const svector<int>& parent_feature_indices =
                        parent.triple_feature_indices;
    size_t parent_feature_count = parent_feature_indices.size();

    distree_node_t& cur_child = tree.m_node_array[child_node_index];
#ifdef USE_MAX_NORM
    BBox child_3x_bbox = cur_child.m_bound;
    vec3f pt_center = child_3x_bbox.Center();
    child_3x_bbox.Expand(child_3x_bbox.MaximumExtentf());
#else
    vec3f pt_center = cur_child.m_bound.Center();
#endif
    
    float child_radius = cur_child.m_bound.BoxRadius();//cur_child.m_bound.SphericalRadius();//
//    int   child_id = tree.compute_local_child_key(cur_child);
//     svector<int>& parent_complementary_features =
//                         parent.complementary_features[child_id];
    bool is_leaf = true;

    for (size_t i = 0; i <  parent_feature_count; i++)
    {
        //scan parent triple features.
        D_ASSERT_P(tree.m_myid, parent_feature_indices[i] < size_feature(),
            "Memory overflow", DERROR_ERR_MEM);
        triangle_t& cur_triangle = m_features_triangle[parent_feature_indices[i]];

        if(cur_triangle.triangle_overlap_sphere(pt_center, child_radius * 3))
        {//record parent's feature that is within 3x BBox of cur_child_index.
            /*
            if (parent.level==2 && parent.key == 132120576)
            {
                printf("Abnormal cut!\n feature_indices=%d, \n pt_center=(%f,%f,%f)\
 child_radius=%f\ntriangle(%f,%f,%f),(%f,%f,%f),(%f,%f,%f)"
                        , parent_feature_indices[i]
                        , pt_center.x(), pt_center.y(), pt_center.z()
                        , child_radius
                        , cur_triangle.v[0].x(), cur_triangle.v[0].y(), cur_triangle.v[0].z()
                       ,cur_triangle.v[1].x(), cur_triangle.v[1].y(), cur_triangle.v[1].z()
                       ,cur_triangle.v[2].x(), cur_triangle.v[2].y(), cur_triangle.v[2].z()
                       ,cur_triangle.vn[0].x(), cur_triangle.vn[0].y(), cur_triangle.vn[0].z()
                       ,cur_triangle.vn[1].x(), cur_triangle.vn[1].y(), cur_triangle.vn[1].z()
                       ,cur_triangle.vn[2].x(), cur_triangle.vn[2].y(), cur_triangle.vn[2].z()
                        );
                fflush(stdout);
            }
            */
            cur_child.triple_feature_indices.push_back(parent_feature_indices[i]);
            
        }
        /****not to use complementary_feature array, it's not working rightnow.
        else if(cur_triangle.triangle_overlap_sphere(pt_center, child_radius * 5))
        {//if current triangle is outside of 3x range of cur_child node, then
        //add the index of current triangle in parent's complementary_features[]
        //if the triangle is within 5x range of cur_child_node.
        parent_complementary_features.push_back(parent_feature_indices[i]);
        }
        */
    }//end of for
    is_leaf = cur_child.triple_feature_indices.empty();
    /****not to use complementary_feature array, it's not working right now.
    if (cur_child.triple_feature_indices.empty())
    {
        parent_complementary_features = parent.triple_feature_indices;
    }
    */
    //Final check if cur_child is leaf.
    //no matter it contains feature or not. Once its extend is < 2.0f
    //or it is at max_depth_id level, it is a leaf node.
    if (   cur_child.m_bound.MaximumExtentf() < tree.min_extend*2.0f
        || parent.level == tree.max_depth_id-1){
            is_leaf = DTRUE;
    }

    D_ASSERT_P(tree.get_processor_id(),
        cur_child.triple_feature_indices.size() <= parent_feature_indices.size(),
        "Memory overflow",
        DERROR_ERR_MEM);
    //Compact cur_child->voxel_indices array.
    //svector<int>(cur_child.triple_feature_indices).swap(cur_child.triple_feature_indices);
    cur_child.triple_feature_indices.shrink_to_fit();
    /****not to use complementary_feature array, it's not working right now.
    parent_complementary_features.shrink_to_fit();
    */
    return is_leaf;
}

float feature_geometry_t::compute_vertex_feature_distance( distree_vertex_t& cur_vertex, int feature_index,
                                                            distree_t& tree )
{
    float min_dist;
    D_ASSERT_P(tree.m_myid, feature_index < m_features_triangle.size(),
        "triangle index out of bounds", DERROR_ERR_MEM);
    min_dist = m_features_triangle[feature_index].vertex_triangle_distance(cur_vertex.coords);
    return min_dist;
}

float feature_geometry_t::compute_vertex_feature_distance( vec3f& coords, int feature_index, distree_t& tree )
{
    float min_dist;
    D_ASSERT_P(tree.m_myid, feature_index < m_features_triangle.size(),
                "triangle index out of bounds", DERROR_ERR_MEM);
    min_dist = m_features_triangle[feature_index].vertex_triangle_distance(coords);
    return min_dist;
}

color_t feature_geometry_t::contain_local_geometry( const vec3i& pxyz, const vec3i& np_xyz )
{
/*vec3i*/vec3f globle_bbox(m_pobj_model->obj_model_get_global_bbox().getDimension());//global_feature_size(global_domain_size);//
    vec3f local_block_min = BLOCK_LOW_FLOAT3D(pxyz, np_xyz, globle_bbox)+m_pobj_model->obj_model_get_global_bbox().pMin;
    vec3f local_block_max = BLOCK_HIGH_FLOAT3D(pxyz, np_xyz, globle_bbox)+m_pobj_model->obj_model_get_global_bbox().pMin;//vec3f(BLOCK_HIGH_3D(pxyz, np_xyz, globle_bbox))+m_obj_model.m_bbox.pMin;
    BBox local_box(local_block_min, local_block_max);
    //vec3i local_size = BLOCK_SIZE_3D(pxyz, np_xyz, globle_bbox);
    //Check every triangle if it overlaps with local bbox.
    vec3i* triangles = (vec3i*)(m_pobj_model->triangle_tab);
      int  ntriangle = m_pobj_model->triangle_num;
    vec3f* vertices  = (vec3f*)(m_pobj_model->vertex_tab);
    //int nvertices = m_obj_model.vertex_num;
    int idxBase = 1;
    int nElements = m_pobj_model->vtx_normal_num > 0 ? 2 : 1;
    for (int iTria=0 ; iTria < ntriangle ; iTria++)
    {//for each triangle
        vec3i indices = triangles[iTria*nElements];
        for (int j=0 ; j < 3 ; j++){
            //for each vertex.
            vec3f& cur_vertex = vertices[indices.v(j) - idxBase];
            if (local_box.Overlap(cur_vertex))
            {
                return BLACK;
            }
        }
    }
    //None of the triangle is contained in the local block.
    return WHITE;
}

