#include "feature_global_leaves_t.h"
#include "distree_t.h"
#include "parallel_test.h"
#include "hpgv_utiltiming.h"
using namespace lily;

feature_global_leaves_t::feature_global_leaves_t(void)
{
}

feature_global_leaves_t::~feature_global_leaves_t(void)
{
}

void feature_global_leaves_t::initial_distree_feature_array( distree_t &tree )
{
    int iLeaf=0;
    for (int in = 0 ; in < tree.m_node_array.size(); in++)
    {//create global_leaf_block_feature that corresponds to each global leaf nodes.
        if (tree.m_node_array[in].isLeaf)
        {
            distree_node_t& cur_node = tree.m_node_array[in];
            m_assigned_global_tree_leaf_nodes.push_back(distree_global_leaf_t());
            distree_global_leaf_t& global_leaf = m_assigned_global_tree_leaf_nodes.back();
            global_leaf.m_bound = cur_node.m_bound;
            global_leaf.processor_indices.clone(cur_node.processor_indices);
            global_leaf.processor_3x_indices.clone(cur_node.processor_3x_indices);
        }
    }
    for (int in = 0 ; in < tree.m_assigned_local_leaf_indices.size(); in++)
    {
        distree_node_t& cur_node = tree.m_node_array[tree.m_assigned_local_leaf_indices[in]];
        if (tree.m_myid == 0)
        {
            fprintf(stderr, "PE%d global leaf node[%d].key level=%d***\n"
                , tree.m_myid
                , iLeaf
                , cur_node.key
                , cur_node.level
                );
            iLeaf++;
        }
    }

    fflush(stderr);
}

size_t feature_global_leaves_t::size_feature() const
{
    return m_assigned_global_tree_leaf_nodes.size();
}

void* feature_global_leaves_t::data()
{
    return data_ptr(m_assigned_global_tree_leaf_nodes);
}

bool feature_global_leaves_t::collect_feature( size_t child_node_index, size_t parent_node_index,
                                                distree_t& tree, bool is_profiling/*=false */ )
{
    //compute child's 3xbbox
    bool is_leaf = true;
    distree_node_t& cur_node	= tree.m_node_array[child_node_index];

    if (cur_node.isLeaf && tree.m_is_building_global_tree)
    {//global leaf nodes are consider as features, so we don't do anything to them.
//      return is_leaf;//
    }

    distree_node_t& parent_node = tree.m_node_array[parent_node_index];
    BBox cur_node_box_3x = cur_node.get_triple_bbox();
    float tolarence = 0.001f;
//  if (tree.m_is_building_global_tree)
    {
        cur_node.processor_indices.clear();
        cur_node.processor_3x_indices.clear();
        cur_node.isoblock_3x_indices.clear();
        cur_node.isoblock_overlap_indices.clear();
    }

    if (is_profiling){
        HPGV_TIMING_BEGIN(parallel_test::T_g_isoblock);
    }
    for (int inode = 0 ; inode < parent_node.isoblock_3x_indices.size() ; inode++)
    {
        int assigned_leaf_index = parent_node.isoblock_3x_indices[inode];
        D_ASSERT_P(tree.m_myid,  assigned_leaf_index < tree.m_node_array.size(),
                    "Memory overflow\n", DERROR_ERR_MEM);

        //distree_node_t& cur_global_tree_leaf = tree.m_node_array[assigned_leaf_index];
        distree_global_leaf_t& assiged_global_tree_leave =
                    m_assigned_global_tree_leaf_nodes[assigned_leaf_index];

//         if (tree.m_is_building_global_tree)
//         {
//             D_ASSERT_P(tree.m_myid,  cur_global_tree_leaf.isLeaf,
//                 "feature_global_leaves_t::collect_feature cur_leaf is not leaf !!\n", DERROR_ERR_MEM);
//         }

        if (cur_node.m_bound.Overlap(assiged_global_tree_leave.m_bound, tolarence))
        {
            cur_node.isoblock_overlap_indices.push_back(assigned_leaf_index, is_profiling);
            cur_node.isoblock_3x_indices.push_back(assigned_leaf_index, is_profiling);

            cur_node.processor_indices.push_back(assiged_global_tree_leave.processor_indices, is_profiling);
            cur_node.processor_3x_indices.push_back(assiged_global_tree_leave.processor_indices, is_profiling);
            if (tree.m_myid == 1 && cur_node.key == 31211520)
            {
                fflush(stderr);
            }
        }else if (cur_node_box_3x.Overlap(assiged_global_tree_leave.m_bound, tolarence))
        {
            cur_node.isoblock_3x_indices.push_back(assigned_leaf_index, is_profiling);
            cur_node.processor_3x_indices.push_back(assiged_global_tree_leave.processor_indices);
        }
    }
    if (is_profiling){
        HPGV_TIMING_END(parallel_test::T_g_isoblock);
    }
    cur_node.processor_3x_indices.remove_duplicates(int_compare);
    cur_node.processor_indices.remove_duplicates(int_compare);
    cur_node.isoblock_3x_indices.remove_duplicates(int_compare);
    cur_node.isoblock_overlap_indices.remove_duplicates(int_compare);
    cur_node.bHas_dirty_proc_3x = tree.node_has_dirty_proc_3x(cur_node);
    cur_node.processor_indices.shrink_to_fit(is_profiling);
    cur_node.processor_3x_indices.shrink_to_fit(is_profiling);
    cur_node.isoblock_3x_indices.shrink_to_fit(is_profiling);
    cur_node.isoblock_overlap_indices.shrink_to_fit(is_profiling);
    //Final check if cur_node is leaf.
    //no matter it contains feature or not. Once its extend is < block_size
    //or it is at max_depth_id-1 level, it is a leaf
    if( parent_node.level == tree.max_depth_id-1
        || cur_node.m_bound.MaximumExtentf() < tree.min_extend)
    {
        is_leaf = true;
    }
    return is_leaf;
}

float feature_global_leaves_t::compute_vertex_feature_distance(distree_vertex_t& cur_vertex, int feature_index, distree_t& tree )
{
 //   throw std::exception("The method or operation is not implemented.");
    return 0.0f;
}
float feature_global_leaves_t::compute_vertex_feature_distance(vec3f& coords, int feature_index, distree_t& tree )
{
  //  throw std::exception("The method or operation is not implemented.");
    return 0.0f;
}
