#include "feature_isoblock_t.h"
#include "distree_t.h"
#include "bitmap_t.h"
#include "mathtool.h"
#include "parallel_test.h"
#include "hpgv_utiltiming.h"
//#define MPI_INCLUDED 1
//#include <DError.h>
using namespace lily;


feature_isoblock_t::feature_isoblock_t( const bitmap_t* bits )
    :bitmap(bits)
{
}

feature_isoblock_t::~feature_isoblock_t( void )
{
}

void feature_isoblock_t::initial_distree_feature_array( distree_t &tree )
{
    //tree.m_center = tree.global_feature_size * 0.5f;
    //Compact all black isoblocks.
    size_t mapsize = bitmap->size;
    int isoblock_count = bitmap->nbits;
    int pe_id = 0;
    m_isoblocks.reserve(mapsize);
    vec3i np_xyz = tree.get_processor_layout();
    vec3f global_feature_size = tree.m_global_feature_size;
    vec3f center = tree.m_global_feature_size*0.5f;
//    cout<<"center="<<center;
    if (isoblock_count > 0){
        //isoblocks = (distree_isoblock_t*) calloc(isoblock_count, sizeof(distree_isoblock_t));
        for (size_t i=0; i < mapsize ; i++){
            //for each chunk of 8bits
            uint8_t chunk = bitmap->data[i];
            for (int j=0; j < 8 && pe_id < tree.m_groupsize ; j++, pe_id++){
                //for each bit~processor
                    //convert pe_id to x,y,z coords
                    vec3i pxyz = vec3i::index1D_to_index3D(pe_id, np_xyz);
                    distree_isoblock_t cur_isoblock;
                    cur_isoblock.pe_id = pe_id;
                    cur_isoblock.has_feature = false;
                    //isoblocks BBox
                    cur_isoblock.m_bound.pMin = BLOCK_LOW_FLOAT3D( pxyz, np_xyz, global_feature_size)-center;
                    cur_isoblock.m_bound.pMax = BLOCK_HIGH_FLOAT3D(pxyz, np_xyz, global_feature_size)-center;
                if ((chunk>>j)&0x01)
                {
                    cur_isoblock.has_feature = true;
                    if (pe_id == tree.m_myid){
                        tree.m_has_features = true;
                    }
                }
                m_isoblocks.push_back(cur_isoblock);
                //pe_id++;
            }
        }
    }
    //cout<<"Isoblock="<<m_isoblocks[0].m_bound;
    isoblock_count = m_isoblocks.size();//this isoblock count also consider empty block now.
    char msg[256];
//     sprintf(msg,"k=%d, bitmap.nbits=%d\n",m_isoblocks.size(), bitmap->nbits);
//     D_ASSERT_P(tree.processor_id(), m_isoblocks.size() == bitmap->nbits,
//                msg, DERROR_ERR_MEM);
//     sprintf(msg,"isoblock count=%d, m_groupsize=%d\n",m_isoblocks.size(), tree.m_groupsize);
//     D_ASSERT_P(tree.processor_id(), m_isoblocks.size() == tree.m_groupsize,
//         msg, DERROR_ERR_MEM);
    if (isoblock_count <= 0){
        return ;
    }
    //shrink_to_fit, similar to realloc() function.
    m_isoblocks.shrink_to_fit();
    //svector<distree_isoblock_t>(m_isoblocks).swap(m_isoblocks);
    //----------generate processor list for distree.--------------
    //So that processor_list[i].m_isoblock_indices records the indices of isoblocks
    //that is assigned to processor i.
    tree.m_processor_array.resize(tree.m_groupsize);
    for (int i = 0; i < isoblock_count; i++){
        int pe_id = m_isoblocks[i].pe_id;
        bool has_feature = m_isoblocks[i].has_feature;
        distree_processor_t& processors = tree.m_processor_array[pe_id];
        processors.m_has_features |= has_feature;
        processors.m_isoblock_indices.push_back(i);
    }
    /*
    if (tree.m_myid == 0){
        char msg[128];
        sprintf(msg, "--------PE%d has %d isoblocks(has/no feature) in feature domain.",
            tree.m_myid,tree.m_processor_array[tree.m_myid].m_isoblock_indices.size());
        D_P_MSG(tree.m_myid, tree.m_myid, msg);
    }
    if (tree.m_myid == 0)
    {
        for (int is = 0 ; is < 10; is ++)
        {
            fprintf( stderr, "PE%d -- isoblock[%d].hasfeatures=%d---\n"
                ,tree.m_myid
                ,is
                ,m_isoblocks[is].has_feature
                );
        }
    }
    if (tree.m_myid == 0)
    {
        for (int ip = 0 ; ip < 10; ip ++)
        {
            fprintf( stderr, "PE%d -- m_processor_array[%d].hasfeatures=%d---\n"
                ,tree.m_myid
                ,ip
                ,tree.m_processor_array[ip].m_has_features
                );
        }
    }
    */
}

size_t feature_isoblock_t::size_feature() const
{
    return m_isoblocks.size();
}

bool feature_isoblock_t::collect_feature( size_t child_node_index,
                                          size_t parent_node_index,
                                          distree_t& tree, bool is_profiling/*=false*/ )
{
    //compute child's 3xbbox
    distree_node_t& cur_node	= tree.m_node_array[child_node_index];
    distree_node_t& parent_node = tree.m_node_array[parent_node_index];

    float cur_node_extend = cur_node.m_bound.MaximumExtentf();
    //compute BBox of cur_node's triple.
    BBox cur_node_box_3x(cur_node.m_bound);
    cur_node_box_3x.Expand(cur_node_extend);

    bool  is_leaf = true;

    if (parent_node.isoblock_3x_indices.size() <= 0){
        volatile char* p=NULL;
        *p = 'c';//Upon exception happen, perform suicide.
    }
    D_ASSERT_P(tree.m_myid, (parent_node.isoblock_3x_indices.size() > 0)
        , "feature_isoblock_t::collect_feature():No isoblock in parent's triple, but why you still decompose parent node?",
        DERROR_ERR_MEM);

    //At most hold as much as isoblocks as its parent have.
    cur_node.isoblock_3x_indices.reserve(parent_node.isoblock_3x_indices.size(),is_profiling);
    cur_node.isoblock_overlap_indices.reserve(parent_node.isoblock_overlap_indices.size(),is_profiling);
    //-----------------------------------------------------------
    //Scan isoblock and choose the one that overlaps with cur_node's triple
    //-----------------------------------------------------------
    if (is_profiling){
        HPGV_TIMING_BEGIN(parallel_test::T_g_isoblock);
    }
    float tolarence = 0.001f;
    for (int i = 0; i < parent_node.isoblock_3x_indices.size() ; i++) {
        int cur_isoblock_idx = parent_node.isoblock_3x_indices[i];

        D_ASSERT_P(tree.m_myid,  cur_isoblock_idx < size_feature(),
                    "Memory overflow", DERROR_ERR_MEM);

        distree_isoblock_t &cur_isoblock = m_isoblocks[cur_isoblock_idx];

        //check if isoblock overlap with BBox of cur_node.
        //if yes, it must also overlap with 3x range of cur_node.
//         if (tree.m_myid == 0)
//         {//print the 0th and 1th isoblock and processor
//             if (cur_isoblock_idx >= 0
//                 && cur_isoblock_idx <= 2
//                 )
//             {
//                 fprintf( stderr, "=++= PE%d isoblock[%d].has_feature=%d processor[%d].has_feature=%d=++=\n"
//                     , tree.m_myid
//                     , cur_isoblock_idx
//                     , cur_isoblock.has_feature
//                     , cur_isoblock.pe_id
//                     , tree.m_processor_array[cur_isoblock.pe_id].m_has_features
//                     );
//             }
//         }

        if( cur_isoblock.has_feature &&
            cur_node.m_bound.Overlap(cur_isoblock.m_bound,tolarence))
        {
            //record feature idx that overlaps BBox of cur_node_index.
            cur_node.isoblock_overlap_indices.push_back(cur_isoblock_idx, is_profiling);
            cur_node.processor_indices.push_back(cur_isoblock.pe_id, is_profiling);
            //record feature idx that is within 3x BBox of cur_node_index.
            //!!Do we really need the following line??
            cur_node.isoblock_3x_indices.push_back(cur_isoblock_idx, is_profiling);
            cur_node.processor_3x_indices.push_back(cur_isoblock.pe_id, is_profiling);
//             if (tree.m_myid == 0)
//             {//print the 0th and 1th isoblock and processor
//                 if (cur_isoblock_idx >= 0
//                     && cur_isoblock_idx <= 2
//                     )
//                 {
//                     fprintf( stderr, "=== PE%d isoblock[%d].has_feature=%d processor[%d].has_feature=%d===\n"
//                         , tree.m_myid
//                         , cur_isoblock_idx
//                         , cur_isoblock.has_feature
//                         , cur_isoblock.pe_id
//                         , tree.m_processor_array[cur_isoblock.pe_id].m_has_features
//                         );
//                 }
//            }
        }else if (cur_isoblock.has_feature &&
            cur_node_box_3x.Overlap(cur_isoblock.m_bound,tolarence))
        {
            //record feature that is within 3x BBox of cur_node_index.
            cur_node.isoblock_3x_indices.push_back(cur_isoblock_idx, is_profiling);
            cur_node.processor_3x_indices.push_back(cur_isoblock.pe_id, is_profiling);
//             if (tree.m_myid == 0)
//             {//print the 0th and 1th isoblock and processor
//                 if (cur_isoblock_idx >= 0
//                     && cur_isoblock_idx <= 2
//                     )
//                 {
//                     fprintf( stderr, "==3x= PE%d isoblock[%d].has_feature=%d processor[%d].has_feature=%d ==3x=\n"
//                         , tree.m_myid
//                         , cur_isoblock_idx
//                         , cur_isoblock.has_feature
//                         , cur_isoblock.pe_id
//                         , tree.m_processor_array[cur_isoblock.pe_id].m_has_features
//                         );
//                 }
//             }
        }
    }
    if (is_profiling){
        HPGV_TIMING_END(parallel_test::T_g_isoblock);
    }
    if (is_profiling){
        HPGV_TIMING_BEGIN(parallel_test::T_g_proc);
    }
    /*
    //At most hold as much as processor_indices as its parent have.
    cur_node.processor_indices.reserve(parent_node.processor_indices.size(),is_profiling);
    cur_node.processor_3x_indices.reserve(parent_node.processor_3x_indices.size(),is_profiling);
    for (int i = 0 ; i < parent_node.processor_3x_indices.size() ; i++){
        //if proc's element blocks intersect child bbox, add proc into
        //child.m_processors_array.
        distree_processor_t& proc = tree.m_processor_array[parent_node.processor_3x_indices[i]];
        for (int j = 0 ; j < proc.m_isoblock_indices.size(); j++){
            int isoblock_idx = proc.m_isoblock_indices[j];
            distree_isoblock_t &cur_isoblock = m_isoblocks[isoblock_idx];
            if (cur_isoblock.m_bound.Overlap(cur_node.m_bound, tolarence))
            {
                cur_node.processor_indices.push_back(parent_node.processor_3x_indices[i], is_profiling);
                cur_node.processor_3x_indices.push_back(parent_node.processor_3x_indices[i], is_profiling);
            }else if (cur_isoblock.m_bound.Overlap(cur_node_box_3x, tolarence))
            {
                cur_node.processor_3x_indices.push_back(parent_node.processor_3x_indices[i], is_profiling);
            }
        }
    }
    */
    if (is_profiling){
        HPGV_TIMING_END(parallel_test::T_g_proc);
    }
    //shrink_to_fit
    cur_node.processor_indices.shrink_to_fit(is_profiling);
    cur_node.processor_3x_indices.shrink_to_fit(is_profiling);
    cur_node.isoblock_3x_indices.shrink_to_fit(is_profiling);
    cur_node.isoblock_overlap_indices.shrink_to_fit(is_profiling);
    cur_node.bHas_dirty_proc_3x = tree.node_has_dirty_proc_3x(cur_node);//
    if (!cur_node.processor_3x_indices.empty()
        && !cur_node.bHas_dirty_proc_3x)
    {
        fprintf(stderr, "+++++PE%d abnormal node.key=%d .level=%d +++++\n"
            , tree.m_myid
            , cur_node.key
            , cur_node.level
            );
    }
    if (tree.m_myid == 0
        && cur_node.key == 0
        && cur_node.level == 3
        //&& !cur_node.bHas_dirty_proc_3x
        )
    {
        fprintf(stderr, "+++++PE%d abnormal node.key=%d .level=%d proc3x.size=%d has3xDirtProc=%d+++++\n"
            , tree.m_myid
            , cur_node.key
            , cur_node.level
            , cur_node.processor_3x_indices.size()
            , cur_node.bHas_dirty_proc_3x
            );
    }
    fflush(stderr);

    //Final check if cur_node is leaf.
    //no matter it contains feature or not. Once its extend is < block_size
    //or it is at max_depth_id-1 level, it is a leaf
    if( parent_node.level == tree.max_depth_id-1
        || cur_node_extend < tree.min_extend)
    {
        is_leaf = true;
    }
    //cur_node->isLeaf = is_leaf;
    return is_leaf;
}

float feature_isoblock_t::compute_vertex_feature_distance( distree_vertex_t& cur_vertex,
                                                    int feature_index, distree_t& tree )
{
    //throw std::exception("The method or operation is not implemented.");
    return 0.0f;
}

float feature_isoblock_t::compute_vertex_feature_distance( vec3f& coords,
                                                    int feature_index, distree_t& tree )
{
     //throw std::exception("The method or operation is not implemented.");
    return 0.0f;
}

void feature_isoblock_t::update_tree( distree_t& tree)
{
    distree_node_t* cur_node = &(tree.m_node_array[0]);
    cur_node->isoblock_3x_indices.clear();
    cur_node->isoblock_3x_indices.resize(size_feature());
    cur_node->isoblock_overlap_indices.clear();
    cur_node->isoblock_overlap_indices.resize(size_feature());
    int count = cur_node->isoblock_3x_indices.size();
    for (int i = 0 ;i < count; i++)
    {
        cur_node->isoblock_3x_indices[i] = i;
        cur_node->isoblock_overlap_indices[i] = i;
    }
    //processor_indice and processor_3x_indice stay the same.
    int parent_start = 0;
    int parent_end = 1; 
//    while (parent_start < m_node)
    {
    }
    
}
