#ifndef _MATHTOOL_H_
#define _MATHTOOL_H_
#include <stdint.h>
#include "vec3f.h"
#include "vec3i.h"

namespace lily{
#define SQUARE(x)\
    ((x)*(x))

#define BLOCK_LOW(id, p, n) \
    ((uint64_t)(((float)(id)) /(p) * (n)))

#define BLOCK_HIGH(id, p, n) \
    (BLOCK_LOW((id) + 1, p, n) - 1)

#define BLOCK_SIZE(id, p, n) \
    (BLOCK_LOW((id) + 1, p, n) - BLOCK_LOW(id, p, n))

inline vec3i BLOCK_HIGH_3D(const vec3i &id3d,const vec3i &procDim,const vec3i &domainDim)
{
    return vec3i(
        BLOCK_HIGH(id3d.x(), procDim.x(), domainDim.x()),
        BLOCK_HIGH(id3d.y(), procDim.y(), domainDim.y()),
        BLOCK_HIGH(id3d.z(), procDim.z(), domainDim.z())
        );
}
inline vec3i BLOCK_LOW_3D(const vec3i &id3d,const vec3i &procDim,const vec3i &domainDim)
{
    return vec3i(
        BLOCK_LOW(id3d.x(), procDim.x(), domainDim.x()),
        BLOCK_LOW(id3d.y(), procDim.y(), domainDim.y()),
        BLOCK_LOW(id3d.z(), procDim.z(), domainDim.z())
        );
}
inline vec3i BLOCK_SIZE_3D(const vec3i &id3d,const vec3i &procDim,const vec3i &domainDim)
{
    return vec3i(
        BLOCK_SIZE(id3d.x(), procDim.x(), domainDim.x()),
        BLOCK_SIZE(id3d.y(), procDim.y(), domainDim.y()),
        BLOCK_SIZE(id3d.z(), procDim.z(), domainDim.z())
        );
}

#define BLOCK_LOW_FLOAT(id, p, n) \
    ((float)(((float)(id)) /(p) * (n)))

#define BLOCK_HIGH_FLOAT(id, p, n) \
    (BLOCK_LOW_FLOAT((id) + 1, p, n))

#define BLOCK_SIZE_FLOAT(id, p, n) \
    (BLOCK_LOW_FLOAT((id) + 1, p, n) - BLOCK_LOW_FLOAT(id, p, n))

inline vec3f BLOCK_HIGH_FLOAT3D(const vec3i &id3d,const vec3i &procDim,const vec3f &domainDim)
{
    return vec3f(
        BLOCK_HIGH_FLOAT(id3d.x(), procDim.x(), domainDim.x()),
        BLOCK_HIGH_FLOAT(id3d.y(), procDim.y(), domainDim.y()),
        BLOCK_HIGH_FLOAT(id3d.z(), procDim.z(), domainDim.z())
        );
}
inline vec3f BLOCK_LOW_FLOAT3D(const vec3i &id3d,const vec3i &procDim,const vec3f &domainDim)
{
    return vec3f(
        BLOCK_LOW_FLOAT(id3d.x(), procDim.x(), domainDim.x()),
        BLOCK_LOW_FLOAT(id3d.y(), procDim.y(), domainDim.y()),
        BLOCK_LOW_FLOAT(id3d.z(), procDim.z(), domainDim.z())
        );
}
inline vec3f BLOCK_SIZE_FLOAT3D(const vec3i &id3d,const vec3i &procDim,const vec3f &domainDim)
{
    return vec3f(
        BLOCK_SIZE_FLOAT(id3d.x(), procDim.x(), domainDim.x()),
        BLOCK_SIZE_FLOAT(id3d.y(), procDim.y(), domainDim.y()),
        BLOCK_SIZE_FLOAT(id3d.z(), procDim.z(), domainDim.z())
        );
}
//Return true if all components are none-zero
inline bool allTrue(const vec3f& v3){
    return (v3[0]!=0.0f && v3[1]!=0.0f && v3[2]!=0.0f);
}
inline bool allTrue(const vec3i& v3){
    return (v3[0] && v3[1] && v3[2]);
}
//Return true if any component is none-zero
inline bool anyTrue(const vec3f& v3){
    return (v3[0]!=0.0f || v3[1]!=0.0f || v3[2]!=0.0f);
}
inline bool anyTrue(const vec3i& v3){
    return (v3[0] || v3[1] || v3[2]);
}
inline float fminf( float a, float b )
{
    return a < b ? a : b;
}

inline float fmaxf(float a, float b)
{
    return a > b ? a : b;
}

inline int maxi(int a, int b)
{
    return a > b ? a : b;
}

inline int mini(int a, int b)
{
    return a < b ? a : b;
}
inline int sign(float a)
{
    return a > 0 ? 1 : ( a < 0 ? -1 : 0);
}
//Reference:http://graphics.stanford.edu/~seander/bithacks.html
//return sign of a, -1,0 or +1.
inline int _sign(float a)
{
    //return (a != 0) | (((int&)a) >> (sizeof(float)*CHAR_BIT-1));
    return (a > 0.0f) - (a < 0.0f);
}
inline int _sign(int a)
{
    //return (a != 0) | (a >> (sizeof(int)*CHAR_BIT-1));
    return (a > 0) - (a < 0);
}
//if a < 0 then 0, else 1
inline int nonnegative(int a)
{
    return 1^(((unsigned int)a) >> (sizeof(int)*CHAR_BIT-1) );
}
inline int nonnegative(float a)
{
    return 1^(((unsigned int&)a) >> (sizeof(float)*CHAR_BIT-1) );
}
//if x,y has opposite signs
inline bool opposigns(int x, int y){
    return (x^y)<0;
}
inline bool opposigns(float x, float y){
    return ((int&)x^(int&)y)<0;
}
//compute integer absolute value w/o branching.
inline int abs_wob(int v){
    int const mask = v >> (sizeof(int)*CHAR_BIT-1);
    return (v^mask)-mask;
}
inline float lerp(float a, float b, float t)
{
    return a + t*(b-a);
}
inline vec3f lerp(const vec3f &a, const vec3f &b, float t)
{
    return a + t*(b-a);
}
inline vec3f slerp(const vec3f& v1, const vec3f& v2, float t)
{
	float dot = v1.dot(v2);
	//Numerical error is finicky, the inner product of two normalized vectors 
	//in parallel is not guaranteed to be within [0,1], clamping is required.
	dot = fminf(dot,1.0f);
	float theta = acosf(dot);

	if (abs(theta) < 1E-3)
	{//Degenerated case.
		return lerp(v1, v2, t);
	}else if (abs(theta-PI) < 1E-3)
	{//Degenerated case.
	  //if theta is close to PI, set it to 179 degree.
		theta = 3.124087f;//179 degree
	}
	float invSinTheta = 1.0f/sinf(theta);
	return invSinTheta*(sinf((1.0f-t)*theta)*v1 + sinf(t*theta)*v2);
}
//************************************************************************
// Method:    tlerp
// Trilinearly interpolate the sampling value in a box.
//   /|\z
//    | [6]-------[7]
//    | /|        /|
//    |/ | /y    / |
//   [2]--/----[3] |
//    |[4]------|--[5]
//    | /       | /
//   [0]-------[1]---->x
// Returns:   interpolate value
// Parameter: const float v[8]: The scalar value on the eight vertices
// Parameter: const vec3f txyz: The offset[0,1] of the sampling point with respect to vertex[0].
//************************************************************************
inline float tlerp(const float v[8], const vec3f txyz)
{
    float sx0 = v[0]*(1-txyz.x()) + v[1]*txyz.x();
    float sx1 = v[2]*(1-txyz.x()) + v[3]*txyz.x();
    float sx2 = v[4]*(1-txyz.x()) + v[5]*txyz.x();
    float sx3 = v[6]*(1-txyz.x()) + v[7]*txyz.x();
    //linear interpolation along y axis
    float sy0 = sx0 * (1-txyz.y()) + sx1 * txyz.y();
    float sy1 = sx2 * (1-txyz.y()) + sx3 * txyz.y();
    //linear interpolation along z axis
    return sy0 * (1-txyz.z()) + sy1 * txyz.z();//fabsf(sz0);
}

//return the component-wise min
vec3f fminf(const vec3f &a,const vec3f &b);
//return the component-wise max
vec3f fmaxf(const vec3f &a,const vec3f &b);
//return the component-wise min
vec3i mini(const vec3i &a,const vec3i &b);
//return the component-wise max
vec3i maxi(const vec3i &a,const vec3i &b);
//return the component-wise min
vec3i minvi(const vec3i &a,const vec3i &b);
//return the component-wise max
vec3i maxvi(const vec3i &a,const vec3i &b);
int int_compare(const void *pa,const void *pb);

}
#endif