#ifndef _OBJ_MODEL_H_
#define _OBJ_MODEL_H_
#pragma once
#include "vec3f.h"
#include "BBox.h"
#include "dutility.h"
#include "triangle_t.h"
#include "vertex_triangle_t.h"

class GLVertexbufferf;
class vertex_triangle_t;

class obj_model_t
{
public:
    float *vertex_tab;        //[x1,y1,z1]..[xn,yn,zn]
    int    vertex_num; //the number of vertex of type __float3

    float *vtx_normal_tab;    //[x1,y1,z1]..[ xn,yn,zn]
    int    vtx_normal_num;//the number of vertex normal of type __float3

    int   *triangle_tab;//[v1,v2,v3,vn1,vn2,vn3]...
    int    triangle_num;//the number of triangle of type __int6

    lily::BBox m_local_bbox;
    int m_baseIndex;//obj file use base-1 index.
#ifdef __USE_GL__
    GLVertexbufferf* m_vbo;
    //float* haha;
#endif
//     vec3f pt_min, pt_max; //bbox of model
//     vec3f center;
public:
    obj_model_t(void);
    ~obj_model_t(void);

    void obj_model_set_baseIndex(int i){m_baseIndex = i;}
    lily::BBox obj_model_get_local_bbox() const;
    virtual lily::BBox obj_model_get_global_bbox() const;
    lily::vec3f obj_model_get_center() const;

    int  obj_model_get_vertex_count() const;

    int  obj_model_get_triangle_count() const;

    int  obj_model_get_normal_count() const;

    //************************************
    // Method:    obj_model_get_triangle_at
    //
    // Return an instance of triangle whose index is idx from obj model.
    // If the method returns false if the specified triangle is degenerated.
    // User discretion is advice when degenerated triangle is return.
    // Returns:   bool: true, if the triangle in normal; false, if the triangle is degenerated.
    // Qualifier: const
    // Description:
    // Parameter: int idx: which triangle to return.
    // Parameter: triangle_t & cur_triangle: output parameter holding the returned triangle.
    //************************************
    bool obj_model_get_triangle_at(int idx, triangle_t& cur_triangle) const;

    void obj_model_save_file(const char* file_name);

virtual  void obj_model_scale_obj(float scale);
virtual  void obj_model_scale_obj_xyz( float sx, float sy, float sz );
virtual  void compute_bbox();

virtual  void obj_model_load_file(const char* file_name);
         void obj_model_load_file2(const char* file_name);
    //set the origin of obj model to one specified by the 2nd param.
    //note that the coordinates of the vertices will be changed accordingly.
virtual  void obj_model_set_origin(lily::vec3f origin);
#ifdef __USE_GL__
    //GL fill mode:GL_LINE, GL_FILL
    virtual  void obj_model_draw(unsigned int fillmode, bool useVBO=true);
#else
    virtual  void obj_model_draw(unsigned int fillmode, bool useVBO=true){};
#endif
    void obj_model_draw_bbox();

    void obj_free(void);
//protected:
    void obj_model_remove_duplicated_vertex(void);
    void obj_model_remove_degenerated_triangle(void);
    void obj_model_print_triangles(void);
    void obj_model_print_vertices(void);
    //convert a set of triangle soups to the
    //vertex_array: a sequence of vertices of triangle
    //myid: the PE id who invoke this method.
    void obj_model_create_from_triangle_soup(svector<vertex_triangle_t>& vertex_array, int myid);
};

#endif

